# Generated by Django 5.0.8 on 2024-09-18 02:54

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0064_remove_conversation_temp_id_alter_conversation_id"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="agent",
            name="avatar",
        ),
        migrations.RemoveField(
            model_name="agent",
            name="public",
        ),
        migrations.AddField(
            model_name="agent",
            name="privacy_level",
            field=models.CharField(
                choices=[("public", "Public"), ("private", "Private"), ("protected", "Protected")],
                default="private",
                max_length=30,
            ),
        ),
        migrations.AddField(
            model_name="entry",
            name="agent",
            field=models.ForeignKey(
                blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, to="database.agent"
            ),
        ),
        migrations.AddField(
            model_name="fileobject",
            name="agent",
            field=models.ForeignKey(
                blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, to="database.agent"
            ),
        ),
        migrations.AlterField(
            model_name="agent",
            name="slug",
            field=models.CharField(max_length=200, unique=True),
        ),
    ]
