input.factChecker_factVerification__UCbIf{width:100%;display:block;padding:12px 20px;border:none;box-sizing:border-box;border-radius:4px;text-align:left;margin:8px auto;font-size:large}div.factChecker_factCheckerContainer__Brqjv{width:75vw;margin:auto}input.factChecker_factVerification__UCbIf:focus{outline:none;box-shadow:0 4px 8px 0 rgba(0,0,0,.2)}div.factChecker_responseText__HzSck{margin:0;padding:0;border-radius:8px}div.factChecker_response__HaFuc{margin-bottom:12px}a.factChecker_titleLink__FA5mt{color:#333;font-weight:700}a.factChecker_subLinks__iaFHn{color:#333;text-decoration:none;font-weight:small;border-radius:4px;font-size:small}div.factChecker_subLinks__iaFHn{display:flex;flex-direction:row;gap:8px;flex-wrap:wrap}div.factChecker_reference___6Fwh{padding:12px;margin:8px;border-radius:8px}footer.factChecker_footer__Tbj3t{width:100%;background:transparent;text-align:left}div.factChecker_reportActions__2_tFq{display:flex;flex-direction:row;gap:8px;justify-content:space-between;margin-top:8px}button.factChecker_factCheckButton__2XPTv{border:none;cursor:pointer;width:100%;border-radius:4px;margin:8px;padding-left:1rem;padding-right:1rem;line-height:1.25rem}button.factChecker_factCheckButton__2XPTv:hover{box-shadow:0 4px 8px 0 rgba(0,0,0,.2)}div.factChecker_spinner__T26Jv{margin:20px;width:40px;height:40px;position:relative;text-align:center;animation:factChecker_sk-rotate__ogES_ 2s linear infinite}div.factChecker_inputFields__LU3cQ{width:100%;display:grid;grid-template-columns:1fr auto;grid-gap:8px}div.factChecker_dot1__n9Bu5,div.factChecker_dot2__HDdBs{width:60%;height:60%;display:inline-block;position:absolute;top:0;border-radius:100%;animation:factChecker_sk-bounce__q4XFt 2s ease-in-out infinite}div.factChecker_dot2__HDdBs{top:auto;bottom:0;animation-delay:-1s}@media screen and (max-width:768px){div.factChecker_factCheckerContainer__Brqjv{width:95vw}}@keyframes factChecker_sk-rotate__ogES_{to{transform:rotate(1turn);-webkit-transform:rotate(1turn)}}@keyframes factChecker_sk-bounce__q4XFt{0%,to{transform:scale(0);-webkit-transform:scale(0)}50%{transform:scale(1);-webkit-transform:scale(1)}}