# Generated by Django 4.2.10 on 2024-05-04 10:10

from typing import List

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0036_publicconversation"),
        ("database", "0039_merge_20240501_0301"),
    ]

    operations: List[str] = []
